@@ -15,13 +15,13 @@
 #
 # See the README file for information on usage and redistribution.
 #
-
+from __future__ import annotations
 
 from . import Image, ImageFile, ImagePalette
 from ._binary import i32be as i32
 
 
-def _accept(prefix):
+def _accept(prefix: bytes) -> bool:
     return len(prefix) >= 4 and i32(prefix) == 0x59A66A95
 
 
@@ -33,7 +33,7 @@ class SunImageFile(ImageFile.ImageFile):
     format = "SUN"
     format_description = "Sun Raster File"
 
-    def _open(self):
+    def _open(self) -> None:
         # The Sun Raster file header is 32 bytes in length
         # and has the following format:
 
@@ -49,6 +49,8 @@ class SunImageFile(ImageFile.ImageFile):
         #         DWORD ColorMapLength;   /* Size of the color map in bytes */
         #     } SUNRASTER;
 
+        assert self.fp is not None
+
         # HEAD
         s = self.fp.read(32)
         if not _accept(s):
@@ -66,21 +68,21 @@ class SunImageFile(ImageFile.ImageFile):
         palette_length = i32(s, 28)
 
         if depth == 1:
-            self.mode, rawmode = "1", "1;I"
+            self._mode, rawmode = "1", "1;I"
         elif depth == 4:
-            self.mode, rawmode = "L", "L;4"
+            self._mode, rawmode = "L", "L;4"
         elif depth == 8:
-            self.mode = rawmode = "L"
+            self._mode = rawmode = "L"
         elif depth == 24:
             if file_type == 3:
-                self.mode, rawmode = "RGB", "RGB"
+                self._mode, rawmode = "RGB", "RGB"
             else:
-                self.mode, rawmode = "RGB", "BGR"
+                self._mode, rawmode = "RGB", "BGR"
         elif depth == 32:
             if file_type == 3:
-                self.mode, rawmode = "RGB", "RGBX"
+                self._mode, rawmode = "RGB", "RGBX"
             else:
-                self.mode, rawmode = "RGB", "BGRX"
+                self._mode, rawmode = "RGB", "BGRX"
         else:
             msg = "Unsupported Mode/Bit Depth"
             raise SyntaxError(msg)
@@ -97,7 +99,7 @@ class SunImageFile(ImageFile.ImageFile):
             offset = offset + palette_length
             self.palette = ImagePalette.raw("RGB;L", self.fp.read(palette_length))
             if self.mode == "L":
-                self.mode = "P"
+                self._mode = "P"
                 rawmode = rawmode.replace("L", "P")
 
         # 16 bit boundaries on stride
